import requests
import ipywidgets as widgets
import geopandas as gpd
import pandas as pd
import geemap as gm
from shapely.geometry import Point
import shapely
from pyproj import Transformer
import json
import os
import numpy as np

# ee.Authenticate()
# ee.Initialize(project="ee-forestplotvariables")


class buffer_coordinates:
    def __init__(self):
        """
        Initializes the buffer_coordinates class and sets up the GUI components.
        Part 1 of buffered coordinates approach which creates circles around points.
        (Part 2 is in buffer_method.py)
        """
        # File Upload
        self.file_upload = widgets.FileUpload(
            accept=".csv, .txt",  # Accepted file extension e.g. '.txt', '.pdf', 'image/*', 'image/*,.pdf'
            multiple=False,  # True to accept multiple files upload else False
        )

        # Dropdown
        url = "https://raw.githubusercontent.com/opengeos/geospatial-data-catalogs/master/gee_catalog.json"
        data = self.fetch_geojson(url)
        data_dict = {item["title"]: item["id"] for item in data if "title" in item}
        self.dropdown = widgets.Dropdown(
            options=data_dict,  # keys shown, values returned
            description="Dataset:",
            disabled=False,
        )
        self.dropdown.observe(self.on_dropdown_change, names="value")

        self.buffer_radius = widgets.FloatText(
            value=10,
            description="Buffer radius (ft):",
            disabled=False,
            display="flex",
            flex_flow="column",
            align_items="stretch",
            style={"description_width": "initial"},
        )

        self.run_button = widgets.Button(
            description="Run Query",
            disabled=False,
            button_style="",  # 'success', 'info', 'warning', 'danger' or ''
            tooltip="Click me",
            icon="rotate right",  # (FontAwesome names without the `fa-` prefix)
        )

        self.output = widgets.Output()

        self.run_button.on_click(self.on_button_clicked)

        self.hbox = widgets.HBox(
            [self.file_upload, self.buffer_radius, self.run_button]
        )

        self.vbox = widgets.VBox([self.hbox, self.output])

    def on_button_clicked(self, b):
        """
        Callback function to handle button click events."""
        with self.output:
            self.output.clear_output()
            print(
                f"GeoJSON file will be saved to Downloads folder under this name:{self.buffer_radius.value}.0.geojson"
            )

            import io

            if self.file_upload.value:
                # For the first file (if multiple=False)
                file_info = self.file_upload.value[0]
                content_bytes = file_info["content"].tobytes()  # file content as bytes
                filename = file_info["name"]
                print(f"Filename: {filename}")
                points = pd.read_csv(io.BytesIO(content_bytes))
                lat_cols = ["lat", "latitude", "y", "LAT", "Latitude", "Lat", "Y"]
                lon_cols = [
                    "lon",
                    "long",
                    "longitude",
                    "x",
                    "LON",
                    "Longitude",
                    "Long",
                    "X",
                ]
                id_cols = ["id", "ID", "plot_ID", "plot_id", "plotID", "plotId"]

                def find_column(possible_names, columns):
                    for name in possible_names:
                        if name in columns:
                            return name
                    # fallback: check case-insensitive match
                    lower_columns = {c.lower(): c for c in columns}
                    for name in possible_names:
                        if name.lower() in lower_columns:
                            return lower_columns[name.lower()]
                    raise ValueError(f"No matching column found for {possible_names}")

                lat_col = find_column(lat_cols, points.columns)
                lon_col = find_column(lon_cols, points.columns)
                id_col = find_column(id_cols, points.columns)
                points = points.rename(
                    columns={lat_col: "LAT", lon_col: "LON", id_col: "plot_ID"}
                )
            else:
                print("Please upload a CSV file.")

            radius_ft = self.buffer_radius.value

            out_dir = os.path.join(os.path.expanduser("~"), "Downloads")
            output_file = f"{radius_ft}.geojson"
            out_path = os.path.join(out_dir, output_file)

            self.obfuscate_points_to_circles(
                data=points,
                radius_feet=radius_ft,
                plot_id_col="plot_ID",
                output_file=out_path,
            )
            print(f"Buffered coordinates GeoJSON saved to {out_path}")

    def on_dropdown_change(self, change):
        """
        Callback function to handle dropdown value changes.
        """
        if change["new"]:
            with self.output:
                self.output.clear_output()
                catalog = "https://raw.githubusercontent.com/opengeos/geospatial-data-catalogs/master/gee_catalog.json"
                data = self.fetch_geojson(catalog)
                data_dict = {item["id"]: item["url"] for item in data if "id" in item}
                change_value = str(change["new"])
                url = data_dict.get(change_value)
                print(f"Selected dataset: {change['new']}")
                print(f"URL: {url}")

    def create_obfuscated_circle(self, point, radius_feet, crs="EPSG:4326"):
        """
        Create a circle polygon (as a shapely geometry) with the given radius in feet,
        where the provided point is randomly located inside the circle (not at the center).
        """
        # Convert radius from feet to meters
        radius_m = radius_feet * 0.3048

        # Project to local UTM for accurate distance calculations
        utm_crs = f"EPSG:326{int((point.x + 180) // 6) + 1}"
        transformer_to_utm = Transformer.from_crs(crs, utm_crs, always_xy=True)
        transformer_to_latlon = Transformer.from_crs(utm_crs, crs, always_xy=True)
        x, y = transformer_to_utm.transform(point.x, point.y)

        # Randomize the point's location within the circle
        # angle = np.random.uniform(0, 2 * np.pi)
        # distance = np.random.uniform(0, radius_m)
        # # Calculate center of the circle so that the point is inside the circle but not at the center
        # center_x = x - distance * np.cos(angle)
        # center_y = y - distance * np.sin(angle)
        center = Point(x, y)  # Point(center_x, center_y)

        # Create the circle at the calculated center
        circle = center.buffer(radius_m, resolution=32)

        # Transform the circle back to WGS84
        circle_latlon = shapely.ops.transform(
            lambda x, y: transformer_to_latlon.transform(x, y), circle
        )
        return circle_latlon

    def obfuscate_points_to_circles(self, data, radius_feet, plot_id_col, output_file):
        """
        Obfuscate points to circles and save as GeoJSON.

        Args:
            data (str, pd.DataFrame, gpd.GeoDataFrame): Input data (GeoJSON, DataFrame, or GeoDataFrame).
            radius_feet (float): Radius of the circle in feet.
            plot_id_col (str): Column name for plot IDs.
            output_file (str): Path to save the output GeoJSON file.
        """
        if isinstance(data, str):
            coordinates = pd.read_csv(data)
            gdf = gpd.GeoDataFrame(
                coordinates,
                geometry=gpd.points_from_xy(coordinates.LON, coordinates.LAT),
                crs="EPSG:4326",  # Directly set CRS during creation
            )
        elif isinstance(data, pd.DataFrame):
            coordinates = data
            gdf = gpd.GeoDataFrame(
                coordinates,
                geometry=gpd.points_from_xy(coordinates.LON, coordinates.LAT),
                crs="EPSG:4326",  # Directly set CRS during creation
            )
        else:
            gdf = data.to_crs(epsg=4326)  # Ensure WGS84

        circles = []
        ids = []
        for idx, row in gdf.iterrows():
            point = row["geometry"]
            circle = self.create_obfuscated_circle(point, radius_feet, crs=gdf.crs)
            circles.append(circle)
            ids.append(row[plot_id_col])
        # Create new GeoDataFrame
        gdf_circles = gpd.GeoDataFrame(
            {plot_id_col: ids, "geometry": circles}, crs=gdf.crs
        )
        geojson_str = gdf_circles.to_json()
        with open(output_file, "w") as f:
            f.write(geojson_str)
        print(f"GeoJSON saved to {output_file}")
        return gdf_circles

    def fetch_geojson(self, url):
        """
        Fetch GeoJSON data from a given URL.

        Args:
            url (str): URL to fetch the GeoJSON data from.
        """
        try:
            response = requests.get(url)
            response.raise_for_status()  # Raises an exception for HTTP errors
            geojson_data = response.json()  # Parse the JSON response
            return geojson_data
        except requests.exceptions.HTTPError as http_err:
            print(f"HTTP error occurred: {http_err}")
        except requests.exceptions.ConnectionError as conn_err:
            print(f"Error connecting to the server: {conn_err}")
        except Exception as err:
            print(f"An error occurred: {err}")
        return None



import ipyleaflet
import ipywidgets as widgets
import geopandas as gpd
import pandas as pd
import geemap as gm

# import os

# This code adds various widgets to the map for user interaction.


class Map(ipyleaflet.Map):
    def __init__(self, center=[37.5, -95], zoom=4, height="600px", **kwargs):
        """Initialize the map with a given center and zoom level.

        Args:
            center (list, optional): Center coordinates of the map. Defaults to [37.5, -95].
            zoom (int, optional): Zoom level of the map. Defaults to 4.
            height (str, optional): Height of the map. Defaults to "600px".
            **kwargs: Additional keyword arguments for ipyleaflet.Map.
        """
        super().__init__(center=center, zoom=zoom, **kwargs)
        self.layout.height = height
        self.scroll_wheel_zoom = True
        self.add_layer_control()
        self.change_basemap()
        self.add_search_marker()
        self.upload_points()
        self.upload_geojson_file()
        self.change_built_in_shapefiles()
        self.on_interaction(self.handle_click)
        self.geojson_button()

    def add_basemap(self, basemap="Esri.WorldImagery"):
        """Add basemap to the map.

        Args:
            basemap (str, optional): Basemap name. Defaults to "Esri.WorldImagery".
        """

        url = eval(f"ipyleaflet.basemaps.{basemap}").build_url()
        layer = ipyleaflet.TileLayer(url=url, name=basemap)
        self.add(layer)

    def add_google_map(self, map_type="ROADMAP"):
        """Add Google Map to the map.

        Args:
            map_type (str, optional): Map type. Defaults to "ROADMAP".
        """
        map_types = {
            "ROADMAP": "m",
            "SATELLITE": "s",
            "HYBRID": "y",
            "TERRAIN": "p",
        }
        map_type = map_types[map_type.upper()]

        url = (
            f"https://mt1.google.com/vt/lyrs={map_type.lower()}&x={{x}}&y={{y}}&z={{z}}"
        )
        layer = ipyleaflet.TileLayer(url=url, name="Google Map")
        self.add(layer)

    def handle_click(self, **kwargs):
        """Handles click events on the map."""
        if kwargs.get("type") == "click":
            self.add_layer(ipyleaflet.Marker(location=kwargs.get("coordinates")))

            print(f"Clicked at: {kwargs.get('coordinates')}")

    def add_layer_control(self):
        """Adds a layer control widget to the map."""
        control = ipyleaflet.LayersControl(position="topright")
        self.add_control(control)

    def add_geojson(self, data, hover_style=None, **kwargs):
        """Adds a GeoJSON layer to the map.

        Args:
            data (str or dict): The GeoJSON data. Can be a file path (str) or a dictionary.
            zoom_to_layer (bool, optional): Whether to zoom to the layer's bounds. Defaults to True.
            hover_style (dict, optional): Style to apply when hovering over features. Defaults to {"color": "yellow", "fillOpacity": 0.2}.
            **kwargs: Additional keyword arguments for the ipyleaflet.GeoJSON layer.

        Raises:
            ValueError: If the data type is invalid.
        """
        import geopandas as gpd

        if hover_style is None:
            hover_style = {"color": "gray", "fillOpacity": 0.2}

        if isinstance(data, str):
            gdf = gpd.read_file(data)
            geojson = gdf.__geo_interface__
        elif isinstance(data, dict):
            geojson = data

        style = {"color": "black", "weight": 1, "opacity": 1}

        if (
            hasattr(self, "current_geojson_layer")
            and self.current_geojson_layer is not None
        ):
            try:
                self.remove_layer(self.current_geojson_layer)
            except Exception:
                pass  # Layer might have already been removed

        layer = ipyleaflet.GeoJSON(
            data=geojson, hover_style=hover_style, style=style, **kwargs
        )
        self.current_geojson_layer = layer
        self.add_layer(layer)

    def add_points(self, data, **kwargs):
        """Adds points to the map.

        Args:
            data (str or dict): The GeoJSON data. Can be a file path (str) or a dictionary.
            **kwargs: Additional keyword arguments for the ipyleaflet.GeoJSON layer.
        """

        # Load data with safety checks
        if isinstance(data, str):
            coordinates = pd.read_csv(data)
            gdf = gpd.GeoDataFrame(
                coordinates,
                geometry=gpd.points_from_xy(coordinates.LON, coordinates.LAT),
                crs="EPSG:4326",  # Directly set CRS during creation
            )
        elif isinstance(data, pd.DataFrame):
            coordinates = data
            gdf = gpd.GeoDataFrame(
                coordinates,
                geometry=gpd.points_from_xy(coordinates.LON, coordinates.LAT),
                crs="EPSG:4326",  # Directly set CRS during creation
            )
        else:
            gdf = data.to_crs(epsg=4326)  # Ensure WGS84

        point_style = {
            "radius": 2,
            "color": "black",
            "dashArray": "2",
            "fillOpacity": 1,
            "fillColor": "red",
            "weight": 1,
        }  # 'color': 'white',

        geo_data = ipyleaflet.GeoData(
            geo_dataframe=gdf,
            point_style=point_style,
        )

        self.add(geo_data)

    def add_shp(self, data, **kwargs):
        """Adds a shapefile to the map.

        Args:
            data (str): The file path to the shapefile.
            **kwargs: Additional keyword arguments for the GeoJSON layer.
        """

        gdf = gpd.read_file(data)
        gdf = gdf.to_crs(epsg=4326)
        geojson = gdf.__geo_interface__
        self.add_geojson(geojson, **kwargs)

    def add_gdf(self, gdf, **kwargs):
        """Adds a GeoDataFrame to the map.

        Args:
            gdf (geopandas.GeoDataFrame): The GeoDataFrame to add.
            **kwargs: Additional keyword arguments for the GeoJSON layer.
        """
        gdf = gdf.to_crs(epsg=4326)
        geojson = gdf.__geo_interface__
        self.add_geojson(geojson, **kwargs)

    def upload_points(self, **kwargs):
        """Uploads points to the map."""

        from ipywidgets import FileUpload, Button, ToggleButton, HBox

        button = Button(
            description="Add Points",
            disabled=False,
            button_style="",  # 'success', 'info', 'warning', 'danger' or ''
            tooltip="Click me",
            size=38,
            icon="plus",  # (FontAwesome names without the `fa-` prefix)
        )

        file_upload = FileUpload(
            accept=".csv",
            multiple=False,
            size=38,
            description="Upload CSV",
            style={"description_width": "initial"},
        )

        close_button = ToggleButton(
            icon="map-pin", value=True, tooltip="Upload points from CSV file"
        )
        close_button.layout = widgets.Layout(width="36px", height="36px")

        hbox = HBox([close_button])

        is_open = [True]

        def on_toggle_change(change):
            """Toggle visibility of dropdown"""
            if is_open[0] and change["new"]:
                hbox.children = (close_button,)
            else:
                hbox.children = (close_button, file_upload, button)

        def on_button_clicked(change):
            """Handles the file upload."""
            import io

            if file_upload.value:
                # For the first file (if multiple=False)
                file_info = file_upload.value[0]
                content_bytes = file_info["content"].tobytes()
                points = pd.read_csv(io.BytesIO(content_bytes))
                self.add_points(points)

        close_button.observe(on_toggle_change, names="value")

        button.on_click(on_button_clicked)

        upload_control = ipyleaflet.WidgetControl(widget=hbox, position="topright")
        self.add_control(upload_control)

    def upload_geojson_file(self, **kwargs):
        """Uploads GeoJSON file to the map."""

        from ipywidgets import FileUpload, Button, ToggleButton, HBox

        button = Button(
            description="Add GeoJSON file",
            disabled=False,
            button_style="",  # 'success', 'info', 'warning', 'danger' or ''
            tooltip="Click me",
            size=38,
            icon="plus",  # (FontAwesome names without the `fa-` prefix)
        )

        file_upload = FileUpload(
            accept=".geojson",
            multiple=False,
            size=38,
            description="Upload GeoJSON",
            style={"description_width": "initial"},
        )

        close_button = ToggleButton(icon="file", value=True, tooltip="Add GeoJSON file")
        close_button.layout = widgets.Layout(width="36px", height="36px")

        hbox = HBox([close_button])

        is_open = [True]

        def on_toggle_change(change):
            """Toggle visibility of dropdown"""
            if change["new"]:
                hbox.children = (close_button,)
            else:
                hbox.children = (close_button, file_upload, button)

        def on_button_clicked(change):
            """Handles the url upload."""
            if file_upload.value and len(file_upload.value) > 0:
                import json
                from ipyleaflet import GeoJSON

                file_info = file_upload.value[0]
                # Get bytes from memoryview
                content_bytes = file_info["content"].tobytes()
                # Decode bytes to string, then load as JSON
                geojson_data = json.loads(content_bytes.decode("utf-8"))
                # Add the GeoJSON layer to the map
                geo_json = GeoJSON(data=geojson_data)
                self.add_layer(geo_json)

        close_button.observe(on_toggle_change, names="value")

        button.on_click(on_button_clicked)

        upload_control = ipyleaflet.WidgetControl(widget=hbox, position="topright")
        self.add_control(upload_control)

    def add_widgets(self):
        """Creates and displays widgets for user interaction."""

        from ipywidgets import jslink
        from ipyleaflet import WidgetControl

        # date_picker = widgets.DatePicker(description="Pick a Date", disabled=False)
        opacity_slider = widgets.FloatSlider(
            value=1,
            min=0,
            max=1.0,
            step=0.01,
            description="Opacity",
            continuous_update=False,
            orientation="horizontal",
            readout=True,
            readout_format=".2f",
            style={"description_width": "initial"},
        )

        basemap_layer = self.layers[1]
        jslink((opacity_slider, "value"), (basemap_layer, "opacity"))
        opacity_control = WidgetControl(widget=opacity_slider, position="topright")

        self.add(opacity_control)
        # self.add(date_control)

    def add_vector(self, data, **kwargs):
        """Adds vector data to the map.

        Args:
            data (str, geopandas.GeoDataFrame, or dict): The vector data. Can be a file path, GeoDataFrame, or GeoJSON dictionary.
            **kwargs: Additional keyword arguments for the GeoJSON layer.

        Raises:
            ValueError: If the data type is invalid.
        """
        import geopandas as gpd

        if isinstance(data, str):
            gdf = gpd.read_file(data)
            self.add_gdf(gdf, **kwargs)
        elif isinstance(data, gpd.GeoDataFrame):
            self.add_gdf(data, **kwargs)
        elif isinstance(data, dict):
            self.add_geojson(data, **kwargs)
        else:
            raise ValueError("Invalid data type")

    def change_basemap(self, **kwargs):
        """Changes the basemap of the map using a dropdown selector."""
        from ipywidgets import Dropdown, ToggleButton, HBox, Layout
        from ipyleaflet import WidgetControl, basemap_to_tiles, basemaps

        # Map dropdown names to actual basemap objects
        BASEMAP_LOOKUP = {
            "OpenStreetMap": basemaps.OpenStreetMap.Mapnik,
            "OpenTopoMap": basemaps.OpenTopoMap,
            "Esri.WorldImagery": basemaps.Esri.WorldImagery,
            "CartoDB.DarkMatter": basemaps.CartoDB.DarkMatter,
            "CartoDB.Positron": basemaps.CartoDB.Positron,
            "CartoDB.PositronNoLabels": basemaps.CartoDB.PositronNoLabels,
            "Esri.WorldTopoMap": basemaps.Esri.WorldTopoMap,
            "Esri.WorldStreetMap": basemaps.Esri.WorldStreetMap,
            "Esri.OceanBasemap": basemaps.Esri.OceanBasemap,
            "Esri.NatGeoWorldMap": basemaps.Esri.NatGeoWorldMap,
            "NASAGIBS.ModisTerraTrueColorCR": basemaps.NASAGIBS.ModisTerraTrueColorCR,
            "NASAGIBS.ModisAquaTrueColorCR": basemaps.NASAGIBS.ModisAquaTrueColorCR,
            "NASAGIBS.ViirsTrueColorCR": basemaps.NASAGIBS.ViirsTrueColorCR,
            "NASAGIBS.ViirsEarthAtNight2012": basemaps.NASAGIBS.ViirsEarthAtNight2012,
        }

        # Create widgets

        toggle = ToggleButton(
            value=True,
            tooltip="Change basemap",
            icon="map-o",
            layout=Layout(width="36px", height="36px"),
        )

        dropdown = Dropdown(
            options=list(BASEMAP_LOOKUP.keys()),
            value="OpenStreetMap",
            description="Basemap:",
            style={"description_width": "initial"},
            layout=Layout(width="250px", height="36px"),
        )

        # Store reference to current basemap layer
        if not hasattr(self, "current_basemap"):
            self.current_basemap = self.layers[0]

        def on_dropdown_change(change):
            """Handle basemap selection changes"""
            if change["new"]:
                new_basemap = basemap_to_tiles(BASEMAP_LOOKUP[change["new"]])
                self.substitute_layer(self.current_basemap, new_basemap)
                self.current_basemap = new_basemap

        # def on_toggle_change(change):
        #     """Toggle visibility of dropdown"""
        #     if change["new"]:
        #         dropdown.layout.visibility = 'visible'  # Show dropdown
        #     else:
        #         dropdown.layout.visibility = 'hidden'   # Hide dropdown

        # Set up event handlers
        dropdown.observe(on_dropdown_change, names="value")

        # Assemble and add control
        widget_box = HBox([toggle])

        is_open = [True]

        def on_toggle_change(change):
            """Toggle visibility of dropdown"""
            if is_open[0] and change["new"]:
                widget_box.children = (toggle,)
            else:
                widget_box.children = (toggle, dropdown)

        toggle.observe(on_toggle_change, names="value")

        control = WidgetControl(widget=widget_box, position="topright")
        self.add_control(control)

    def change_built_in_shapefiles(self, **kwargs):
        """Changes the basemap of the map using a dropdown selector."""
        from ipywidgets import Dropdown, ToggleButton, HBox, Layout
        from ipyleaflet import WidgetControl

        # Map dropdown names to actual basemap objects
        SHAPEFILE_LOOKUP = {
            "None": None,
            "U.S. States": "../data/us-states.json",
            "National Forests": "https://geohub.oregon.gov/api/download/v1/items/b479e4bd7d70439a87e0230c99bddce5/geojson?layers=0",
        }

        # Create widgets
        toggle = ToggleButton(
            value=True,
            tooltip="Add built-in shapefile",
            icon="square-o",
            layout=Layout(width="38px", height="38px"),
        )

        dropdown = Dropdown(
            options=list(SHAPEFILE_LOOKUP.keys()),
            value="None",
            placeholder="Select a shapefile",
            description="Select shapefile:",
            style={"description_width": "initial"},
            layout=Layout(width="250px", height="38px"),
        )

        # Store reference to current basemap layer
        if not hasattr(self, "current_basemap"):
            self.current_basemap = self.layers[0]

        def on_dropdown_change(change):
            """Handle shapefile selection changes"""
            if change["new"]:
                geojson_path = SHAPEFILE_LOOKUP[change["new"]]
                if geojson_path is not None:
                    self.add_geojson(geojson_path, name=change["new"])
                else:
                    # Remove existing shapefile layer if "None" is selected
                    if (
                        hasattr(self, "current_geojson_layer")
                        and self.current_geojson_layer is not None
                    ):
                        try:
                            self.remove_layer(self.current_geojson_layer)
                        except Exception:
                            pass

        # Set up event handlers
        dropdown.observe(on_dropdown_change, names="value")

        # Assemble and add control
        widget_box = HBox([toggle])

        is_open = [True]

        def on_toggle_change(change):
            """Toggle visibility of dropdown"""
            if change["new"]:
                widget_box.children = (toggle,)
            else:
                widget_box.children = (toggle, dropdown)

        toggle.observe(on_toggle_change, names="value")

        control = WidgetControl(widget=widget_box, position="topright")
        self.add_control(control)

    def add_search_marker(self, **kwargs):
        """Adds a search marker to the map."""

        from ipyleaflet import SearchControl, Marker, AwesomeIcon

        search_marker = Marker(
            icon=AwesomeIcon(name="search", marker_color="blue", icon_color="white")
        )

        search_control = SearchControl(
            position="topleft",
            url="https://nominatim.openstreetmap.org/search?format=json&q={s}",
            zoom=10,
            marker=search_marker,
        )

        self.add(search_control)

    def geojson_button(self, **kwargs):
        """Creates a button to add shapefiles."""
        from ipywidgets import Button, Text, ToggleButton, HBox, Layout
        from ipyleaflet import WidgetControl

        # Create widgets
        toggle_close = ToggleButton(
            value=True, tooltip="Add GeoJSON from link", icon="link"
        )

        url = Text(
            placeholder="Type something",
            description="GeoJSON URL:",
            disabled=False,
            style={"description_width": "initial"},
        )

        run_button = Button(
            description="Add GeoJSON",
            disabled=False,
            button_style="",  # 'success', 'info', 'warning', 'danger' or ''
            tooltip="Click me",
            size=38,
            icon="plus",  # (FontAwesome names without the `fa-` prefix)
        )

        toggle_close.layout = Layout(width="36px", height="36px")

        # Assemble and add control
        widget_box = HBox([toggle_close])

        def on_toggle_change(change):
            """Toggle visibility of dropdown"""
            if change["new"]:
                widget_box.children = (toggle_close,)
            else:
                widget_box.children = (toggle_close, url, run_button)

        def on_button_clicked(change):
            """Handles the url upload."""
            if url.value:
                # For the first file (if multiple=False)
                geojson_path = url.value
                self.add_geojson(geojson_path, name=url.value)

        toggle_close.observe(on_toggle_change, names="value")

        run_button.on_click(on_button_clicked)

        control = WidgetControl(widget=widget_box, position="topright")

        self.add_control(control)



#!/usr/bin/env python

"""Tests for common module."""

import unittest
from unittest.mock import patch
from io import StringIO

from skiba.common import hello_world


class TestCommon(unittest.TestCase):
    """Tests for common module functions."""

    def test_hello_world_prints_correct_message(self):
        """Test that hello_world prints 'Hello World!' to stdout."""
        with patch("sys.stdout", new=StringIO()) as fake_output:
            hello_world()
            self.assertEqual(fake_output.getvalue().strip(), "Hello World!")

    def test_hello_world_returns_none(self):
        """Test that hello_world returns None."""
        with patch("sys.stdout", new=StringIO()):
            result = hello_world()
            self.assertIsNone(result)


if __name__ == "__main__":
    unittest.main()



import folium
import folium.plugins


class Map(folium.Map):
    """A custom Map class that extends folium.Map."""

    def __init__(self, center=(0, 0), zoom=2, **kwargs):
        """Initializes the Map object.

        Args:
            center (tuple, optional): The initial center of the map as (latitude, longitude). Defaults to (0, 0).
            zoom (int, optional): The initial zoom level of the map. Defaults to 2.
            **kwargs: Additional keyword arguments for the folium.Map class.
        """
        super().__init__(location=center, zoom_start=zoom, **kwargs)

    def add_geojson(
        self,
        data,
        zoom_to_layer=True,
        hover_style=None,
        **kwargs,
    ):
        """Adds a GeoJSON layer to the map.

        Args:
            data (str or dict): The GeoJSON data. Can be a file path (str) or a dictionary.
            zoom_to_layer (bool, optional): Whether to zoom to the layer's bounds. Defaults to True.
            hover_style (dict, optional): Style to apply when hovering over features. Defaults to {"color": "yellow", "fillOpacity": 0.2}.
            **kwargs: Additional keyword arguments for the folium.GeoJson layer.

        Raises:
            ValueError: If the data type is invalid.
        """
        import geopandas as gpd

        if hover_style is None:
            hover_style = {"color": "yellow", "fillOpacity": 0.2}

        if isinstance(data, str):
            gdf = gpd.read_file(data)
            geojson = gdf.__geo_interface__
        elif isinstance(data, dict):
            geojson = data

        geojson = folium.GeoJson(data=geojson, **kwargs)
        geojson.add_to(self)

    def add_shp(self, data, **kwargs):
        """Adds a shapefile to the map.

        Args:
            data (str): The file path to the shapefile.
            **kwargs: Additional keyword arguments for the GeoJSON layer.
        """
        import geopandas as gpd

        gdf = gpd.read_file(data)
        gdf = gdf.to_crs(epsg=4326)
        geojson = gdf.__geo_interface__
        self.add_geojson(geojson, **kwargs)

    def add_gdf(self, gdf, **kwargs):
        """Adds a GeoDataFrame to the map.

        Args:
            gdf (geopandas.GeoDataFrame): The GeoDataFrame to add.
            **kwargs: Additional keyword arguments for the GeoJSON layer.
        """
        gdf = gdf.to_crs(epsg=4326)
        geojson = gdf.__geo_interface__
        self.add_geojson(geojson, **kwargs)

    def add_vector(self, data, **kwargs):
        """Adds vector data to the map.

        Args:
            data (str, geopandas.GeoDataFrame, or dict): The vector data. Can be a file path, GeoDataFrame, or GeoJSON dictionary.
            **kwargs: Additional keyword arguments for the GeoJSON layer.

        Raises:
            ValueError: If the data type is invalid.
        """
        import geopandas as gpd

        if isinstance(data, str):
            gdf = gpd.read_file(data)
            self.add_gdf(gdf, **kwargs)
        elif isinstance(data, gpd.GeoDataFrame):
            self.add_gdf(data, **kwargs)
        elif isinstance(data, dict):
            self.add_geojson(data, **kwargs)
        else:
            raise ValueError("Invalid data type")

    def add_layer_control(self):
        """Adds a layer control widget to the map."""
        folium.LayerControl().add_to(self)

    def add_split_map(self, left="openstreetmap", right="cartodbpositron", **kwargs):
        """Add a split map to the folium map. Can handle TIF files.

        Args:
            left (str, optional): basemap for left side. Defaults to "openstreetmap".
            right (str, optional): basemap for right side. Defaults to "cartodbpositron".
        """

        from localtileserver import TileClient, get_folium_tile_layer

        # map_types = {
        #     "ROADMAP": "m",
        #     "SATELLITE": "s",
        #     "HYBRID": "y",
        #     "TERRAIN": "p",
        # }

        # map_type = map_types[map_type.upper()]

        # url = (
        #     f"https://mt1.google.com/vt/lyrs={map_type.lower()}&x={{x}}&y={{y}}&z={{z}}"
        # )
        def _create_layer(source, **kwargs):
            if str(source).lower().endswith((".tif", ".tiff")):
                # Create a TileClient for the raster file
                tile_client = TileClient(source)
                # Generate a Folium tile layer from the TileClient
                return get_folium_tile_layer(tile_client, **kwargs)
            else:
                # Create a standard Folium TileLayer for basemaps
                return folium.TileLayer(source, **kwargs)

        layer_right = _create_layer(right, **kwargs)
        layer_left = _create_layer(left, **kwargs)

        sbs = folium.plugins.SideBySideLayers(
            layer_left=layer_left, layer_right=layer_right
        )

        layer_left.add_to(self)
        layer_right.add_to(self)
        sbs.add_to(self)

    def change_basemap(self, **kwargs):
        """Changes the basemap of the map.

        Args:
            **kwargs: Additional keyword arguments for the folium.TileLayer class.
        """
        from ipywidgets import jslink, Dropdown
        from ipyleaflet import WidgetControl

        # Step 4: Create the dropdown
        data_dict = {
            "OpenStreetMap": "openstreetmap",
            "CartoDB Positron": "cartodbpositron",
            "Stamen Terrain": "stamenterrain",
            "Stamen Toner": "stamentoner",
            "Stamen Watercolor": "stamenwatercolor",
        }
        dropdown = Dropdown(
            options=data_dict,  # keys shown, values returned
            description="Dataset:",
            disabled=False,
        )

        basemap_layer = self.layers[1]
        jslink((dropdown, "value"), (basemap_layer, "basemap"))
        dropdown = WidgetControl(widget=dropdown, position="topright")

        self.add(dropdown)

    def on_dropdown_change(change):
        if change["new"]:
            m.add_basemap(change["new"])



import requests
import ipywidgets as widgets
import geopandas as gpd
import pandas as pd
import geemap as gm
import ee
import os

# ee.Authenticate()
# ee.Initialize(project="ee-forestplotvariables")


class buffer_method:
    def __init__(self):
        """
        Initializes the buffer_method class and sets up the GUI components.
        Part 2 of buffered coordinates approach which accesses GEE datasets and extracts median values.
        (Part 1 is in buffer_coordinates.py)
        """
        # File Upload
        self.file_upload = widgets.FileUpload(
            accept=".geojson",  # Accepted file extension e.g. '.txt', '.pdf', 'image/*', 'image/*,.pdf'
            multiple=False,  # True to accept multiple files upload else False
        )
        # Dropdown
        url = "https://raw.githubusercontent.com/opengeos/geospatial-data-catalogs/master/gee_catalog.json"
        data = buffer_method.fetch_geojson(url)
        data_dict = {item["title"]: item["id"] for item in data if "title" in item}
        self.dropdown = widgets.Dropdown(
            options=data_dict,  # keys shown, values returned
            description="Dataset:",
            disabled=False,
        )

        self.start_date = widgets.DatePicker(description="Start Date", disabled=False)
        self.end_date = widgets.DatePicker(description="End Date", disabled=False)

        self.run_button = widgets.Button(
            description="Run Query",
            disabled=False,
            button_style="",  # 'success', 'info', 'warning', 'danger' or ''
            tooltip="Click me",
            icon="check",  # (FontAwesome names without the `fa-` prefix)
        )

        self.output = widgets.Output()

        self.run_button.on_click(self.on_button_clicked)
        self.dropdown.observe(self.on_dropdown_change, names="value")

        self.hbox = widgets.HBox([self.file_upload, self.dropdown])

        self.hbox_bottom = widgets.HBox(
            [self.start_date, self.end_date, self.run_button]
        )
        self.vbox = widgets.VBox([self.hbox, self.hbox_bottom, self.output])

    def on_dropdown_change(self, change):
        """
        Callback function to handle dropdown value changes."""
        if change["new"]:
            with self.output:
                self.output.clear_output()
                catalog = "https://raw.githubusercontent.com/opengeos/geospatial-data-catalogs/master/gee_catalog.json"
                data = buffer_method.fetch_geojson(catalog)
                data_dict = {item["id"]: item["url"] for item in data if "id" in item}
                change_value = str(change["new"])
                url = data_dict.get(change_value)
                print(f"Selected dataset: {change['new']}")
                print(f"URL: {url}")

    def on_button_clicked(self, b):
        """
        Callback function to handle button click events.
        """
        import io

        with self.output:
            self.output.clear_output()
            print(
                f"You entered: {self.dropdown.value}. CSV file will be saved to Downloads folder under this name."
            )

            if self.file_upload.value:
                file_info = self.file_upload.value[0]
                content_bytes = file_info["content"].tobytes()
                filename = file_info["name"]
                print(f"Filename: {filename}")

                # --- Modification: Read GeoJSON from bytes ---
                # Use BytesIO to read the uploaded GeoJSON file
                geojson_buffer = io.BytesIO(content_bytes)
                points = gpd.read_file(geojson_buffer)
            else:
                print("Please upload a GeoJSON file.")
                return

            geedata = self.dropdown.value
            start_date = self.start_date.value
            end_date = self.end_date.value

            self.extract_median_values(
                data=points, geedata=geedata, start_date=start_date, end_date=end_date
            )

    def extract_median_values(self, data, geedata, start_date, end_date, **kwargs):
        """
        Extracts median values from a GEE dataset for the given geometry.

        Args:
            data (str, pd.DataFrame, gpd.GeoDataFrame): Input data (GeoJSON, DataFrame, or GeoDataFrame).
            geedata (str): GEE dataset ID.
            start_date (str): Start date for filtering the dataset.
            end_date (str): End date for filtering the dataset.
            **kwargs: Additional arguments for the GEE dataset.
        """

        url = "https://raw.githubusercontent.com/opengeos/geospatial-data-catalogs/master/gee_catalog.json"
        response = requests.get(url)
        response.raise_for_status()  # Raises an exception for HTTP errors
        geojson_data = response.json()
        data_type = [item["type"] for item in geojson_data if item["id"] == geedata]
        data_str = " ".join(data_type)
        start_date = str(start_date)
        end_date = str(end_date)

        if isinstance(data, str):
            # Assume the file is a GeoJSON or Shapefile with polygons
            gdf = gpd.read_file(data)
            if gdf.crs is None:
                gdf.set_crs("EPSG:4326", inplace=True)
            else:
                gdf = gdf.to_crs("EPSG:4326")
        elif isinstance(data, pd.DataFrame):
            # If you have a DataFrame, it should already have a 'geometry' column with Polygon objects
            # If not, you need to construct it-otherwise, just convert to GeoDataFrame
            gdf = gpd.GeoDataFrame(data, geometry="geometry")
            if gdf.crs is None:
                gdf.set_crs("EPSG:4326", inplace=True)
            else:
                gdf = gdf.to_crs("EPSG:4326")
        else:
            # If already a GeoDataFrame
            if data.crs is None:
                gdf = data.set_crs("EPSG:4326")
            else:
                gdf = data.to_crs("EPSG:4326")

        comp_results = pd.DataFrame()

        for row in gdf.itertuples():
            gdf_row = gdf.iloc[[row.Index]]
            geojson = gdf_row.__geo_interface__
            fc = gm.geojson_to_ee(geojson)

            # Try loading as Image
            if data_str == "image":
                col = ee.Image(geedata)
                col = ee.ImageCollection(col)

            elif data_str == "image_collection":
                col = ee.ImageCollection(geedata)
                # If date filters are provided, apply them

            else:
                raise ValueError("Dataset ID is not a valid Image or ImageCollection.")

            if start_date is None and end_date is None:
                col = col.filterDate(start_date, end_date)
            else:
                pass

            zonal_stats = col.median().reduceRegion(
                reducer=ee.Reducer.median(), geometry=fc.geometry()
            )
            print("Zonal Stats:")
            print(zonal_stats)
            col = zonal_stats.getInfo()
            print(col)

            col = pd.DataFrame.from_dict(col, orient="index")

            print(col)
            print(col.keys())
            print(type(col))

            # col['plot_ID'] = gdf_row['plot_ID'].values[0]
            # col = pd.DataFrame(col)
            comp_results = pd.concat([comp_results, col], axis=0)

        # Load the GEE dataset as an image
        name = geedata
        file_name = name.replace("/", "_")

        out_dir = os.path.join(os.path.expanduser("~"), "Downloads")
        output_file = f"{file_name}.csv"
        out_path = os.path.join(out_dir, output_file)

        # Retrieve data from the image using sampleRegions
        sampled_data = comp_results.to_csv(out_path)
        return sampled_data

        # Try loading as FeatureCollection (convert to raster)

        # Load the GEE dataset as an image
        # geeimage = buffer_method.load_gee_as_image(geedata, start_date, end_date)
        # name = geedata
        # file_name = name.replace("/", "_")

        # out_dir = os.path.join(os.path.expanduser("~"), "Downloads")
        # output_file = f"{file_name}.csv"
        # out_path = os.path.join(out_dir, output_file)

        # # Retrieve data from the image using sampleRegions
        # sampled_data = gm.zonal_statistics(geeimage, fc, out_path, stat_type="MEDIAN")

        # return sampled_data

    def fetch_geojson(url):
        """
        Fetches GeoJSON data from a given URL.

        Args:
            url (str): URL to the GeoJSON file."""
        try:
            response = requests.get(url)
            response.raise_for_status()  # Raises an exception for HTTP errors
            geojson_data = response.json()  # Parse the JSON response
            return geojson_data
        except requests.exceptions.HTTPError as http_err:
            print(f"HTTP error occurred: {http_err}")
        except requests.exceptions.ConnectionError as conn_err:
            print(f"Error connecting to the server: {conn_err}")
        except Exception as err:
            print(f"An error occurred: {err}")
        return None

    def create_dropdown():
        """
        Creates an ipywidgets dropdown menu from a GeoJSON catalog.

        Args:
            url (str, optional): URL to the GeoJSON catalog. Defaults to the Opengeos catalog.

        Returns:
            ipywidgets.Dropdown: A dropdown widget with the names from the catalog.
        """

        url = "https://raw.githubusercontent.com/opengeos/geospatial-data-catalogs/master/gee_catalog.json"

        data = buffer_method.fetch_geojson(url)

        data_dict = {item["title"]: item["id"] for item in data if "title" in item}

        dropdown = widgets.Dropdown(
            options=data_dict,  # keys shown, values returned
            description="Dataset:",
            disabled=False,
        )

        return dropdown

    def add_date_picker():
        """
        Creates a date picker widget for selecting dates.

        Returns:
            ipywidgets.DatePicker: A date picker widget for selecting dates."""
        date_picker = widgets.DatePicker(description="Pick a Date", disabled=False)

        return date_picker

    def load_gee_as_image(dataset_id, start_date=None, end_date=None, **kwargs):
        """
        Loads any GEE dataset (Image, ImageCollection, FeatureCollection) as an ee.Image.
        Optionally filters by start and end date if applicable.

        Parameters:
            dataset_id (str): The Earth Engine dataset ID.
            start_date (str): Optional start date in 'YYYY-MM-DD' format.
            end_date (str): Optional end date in 'YYYY-MM-DD' format.

        Returns:
            ee.Image: The resulting image.
        """
        url = "https://raw.githubusercontent.com/opengeos/geospatial-data-catalogs/master/gee_catalog.json"
        response = requests.get(url)
        response.raise_for_status()  # Raises an exception for HTTP errors
        geojson_data = response.json()
        data_type = [item["type"] for item in geojson_data if item["id"] == dataset_id]
        data_str = " ".join(data_type)
        start_date = str(start_date)
        end_date = str(end_date)
        # Try loading as Image
        if data_str == "image":
            img = ee.Image(dataset_id)
            # If .getInfo() doesn't throw, it's an Image
            # img.getInfo()
            return img
        elif data_str == "image_collection":
            col = ee.ImageCollection(dataset_id)
            # If date filters are provided, apply them
            if start_date is None and end_date is None:
                col = col.filterDate(start_date, end_date)
            else:
                pass
            # Reduce to a single image (e.g., median composite)
            img = col.median()
            return img
        # Try loading as FeatureCollection (convert to raster)
        else:
            # fc_temp = ee.FeatureCollection(dataset_id)
            # if start_date is None and end_date is None:
            #         fc_temp = fc_temp.filterDate(start_date, end_date)
            # # Convert to raster: burn a value of 1 into a new image
            # img = fc_temp.reduceToImage(properties=[], reducer=ee.Reducer.median())
            # img.getInfo()
            # return img
            # or print(f"Dataset must be either an Image or Image Collection")
            raise ValueError("Dataset ID is not a valid Image or ImageCollection.")



"""Main module."""

import os
import ipyleaflet


class Map(ipyleaflet.Map):
    def __init__(self, center=[20, 0], zoom=2, height="600px", **kwargs):

        super().__init__(center=center, zoom=zoom, **kwargs)
        self.layout.height = height
        self.scroll_wheel_zoom = True

    def add_basemap(self, basemap="OpenTopoMap"):
        """Add basemap to the map.

        Args:
            basemap (str, optional): Basemap name. Defaults to "OpenTopoMap".
        """

        url = eval(f"ipyleaflet.basemaps.{basemap}").build_url()
        layer = ipyleaflet.TileLayer(url=url, name=basemap)
        self.add(layer)

    def add_google_map(self, map_type="ROADMAP"):
        """Add Google Map to the map.

        Args:
            map_type (str, optional): Map type. Defaults to "ROADMAP".
        """
        map_types = {
            "ROADMAP": "m",
            "SATELLITE": "s",
            "HYBRID": "y",
            "TERRAIN": "p",
        }
        map_type = map_types[map_type.upper()]

        url = (
            f"https://mt1.google.com/vt/lyrs={map_type.lower()}&x={{x}}&y={{y}}&z={{z}}"
        )
        layer = ipyleaflet.TileLayer(url=url, name="Google Map")
        self.add(layer)

    def add_geojson(
        self,
        data,
        zoom_to_layer=True,
        hover_style=None,
        **kwargs,
    ):
        """Adds a GeoJSON layer to the map.

        Args:
            data (str or dict): The GeoJSON data. Can be a file path (str) or a dictionary.
            zoom_to_layer (bool, optional): Whether to zoom to the layer's bounds. Defaults to True.
            hover_style (dict, optional): Style to apply when hovering over features. Defaults to {"color": "yellow", "fillOpacity": 0.2}.
            **kwargs: Additional keyword arguments for the ipyleaflet.GeoJSON layer.

        Raises:
            ValueError: If the data type is invalid.
        """
        import geopandas as gpd

        if hover_style is None:
            hover_style = {"color": "yellow", "fillOpacity": 0.2}

        if isinstance(data, str):
            gdf = gpd.read_file(data)
            geojson = gdf.__geo_interface__
        elif isinstance(data, dict):
            geojson = data
        layer = ipyleaflet.GeoJSON(data=geojson, hover_style=hover_style, **kwargs)
        self.add_layer(layer)

        if zoom_to_layer:
            bounds = gdf.total_bounds
            self.fit_bounds([[bounds[1], bounds[0]], [bounds[3], bounds[2]]])

    def add_shp(self, data, **kwargs):
        """Adds a shapefile to the map.

        Args:
            data (str): The file path to the shapefile.
            **kwargs: Additional keyword arguments for the GeoJSON layer.
        """
        import geopandas as gpd

        gdf = gpd.read_file(data)
        gdf = gdf.to_crs(epsg=4326)
        geojson = gdf.__geo_interface__
        self.add_geojson(geojson, **kwargs)

    def add_gdf(self, gdf, **kwargs):
        """Adds a GeoDataFrame to the map.

        Args:
            gdf (geopandas.GeoDataFrame): The GeoDataFrame to add.
            **kwargs: Additional keyword arguments for the GeoJSON layer.
        """
        gdf = gdf.to_crs(epsg=4326)
        geojson = gdf.__geo_interface__
        self.add_geojson(geojson, **kwargs)

    def add_vector(self, data, **kwargs):
        """Adds vector data to the map.

        Args:
            data (str, geopandas.GeoDataFrame, or dict): The vector data. Can be a file path, GeoDataFrame, or GeoJSON dictionary.
            **kwargs: Additional keyword arguments for the GeoJSON layer.

        Raises:
            ValueError: If the data type is invalid.
        """
        import geopandas as gpd

        if isinstance(data, str):
            gdf = gpd.read_file(data)
            self.add_gdf(gdf, **kwargs)
        elif isinstance(data, gpd.GeoDataFrame):
            self.add_gdf(data, **kwargs)
        elif isinstance(data, dict):
            self.add_geojson(data, **kwargs)
        else:
            raise ValueError("Invalid data type")

    def handle_click(self, **kwargs):
        import ipywidgets as widgets

        if kwargs.get("type") == "click":
            self.add_layer(widgets.Marker(location=kwargs.get("coordinates")))

    def add_layer_control(self):
        """Adds a layer control widget to the map."""
        control = ipyleaflet.LayersControl(position="topright")
        self.add_control(control)

    def add_raster(self, filepath, **kwargs):
        """Add GeoTIFF raster to the map.

        Args:
            filepath (_type_): _description_
        """
        from localtileserver import TileClient, get_leaflet_tile_layer

        client = TileClient(filepath)
        tile_layer = get_leaflet_tile_layer(client, **kwargs)

        self.add(tile_layer)
        self.center = client.center()
        self.zoom = client.default_zoom

    def add_image(self, image, bounds=None, **kwargs):
        """Adds an image to the map.

        Args:
            image (str): The file path to the image.
            bounds (list, optional): The bounds for the image. Defaults to None.
            **kwargs: Additional keyword arguments for the ipyleaflet.ImageOverlay layer.
        """

        if bounds is None:
            bounds = [[-90, -180], [90, 180]]
        overlay = ipyleaflet.ImageOverlay(url=image, bounds=bounds, **kwargs)
        self.add(overlay)

    def add_video(self, video, bounds=None, **kwargs):
        """Adds a video to the map.

        Args:
            video (str): The file path to the video.
            bounds (list, optional): The bounds for the video. Defaults to None.
            **kwargs: Additional keyword arguments for the ipyleaflet.VideoOverlay layer.
        """

        if bounds is None:
            bounds = [[-90, -180], [90, 180]]
        overlay = ipyleaflet.VideoOverlay(url=video, bounds=bounds, **kwargs)
        self.add(overlay)

    def add_wms_layer(
        self, url, layers, format="image/png", transparent=True, **kwargs
    ):
        """Adds a WMS layer to the map.

        Args:
            url (str): The WMS service URL.
            layers (str): The layers to display.
            **kwargs: Additional keyword arguments for the ipyleaflet.WMSLayer layer.
        """
        layer = ipyleaflet.WMSLayer(
            url=url, layers=layers, format=format, transparent=transparent, **kwargs
        )

        self.add(layer)

    def change_basemap(self, **kwargs):
        """Changes the basemap of the map using a dropdown selector."""
        from ipywidgets import Dropdown, ToggleButton, HBox, Layout
        from ipyleaflet import WidgetControl, basemap_to_tiles, basemaps

        # Map dropdown names to actual basemap objects
        BASEMAP_LOOKUP = {
            "OpenStreetMap": basemaps.OpenStreetMap.Mapnik,
            "OpenTopoMap": basemaps.OpenTopoMap,
            "Esri.WorldImagery": basemaps.Esri.WorldImagery,
            "CartoDB.DarkMatter": basemaps.CartoDB.DarkMatter,
        }

        # Create widgets
        toggle = ToggleButton(
            value=True,
            tooltip="Toggle basemap selector",
            icon="map",
            layout=Layout(width="38px", height="38px"),
        )

        dropdown = Dropdown(
            options=list(BASEMAP_LOOKUP.keys()),
            value="OpenStreetMap",
            description="Basemap:",
            style={"description_width": "initial"},
            layout=Layout(width="250px", height="38px"),
        )

        # Store reference to current basemap layer
        if not hasattr(self, "current_basemap"):
            self.current_basemap = self.layers[0]

        def on_dropdown_change(change):
            """Handle basemap selection changes"""
            if change["new"]:
                new_basemap = basemap_to_tiles(BASEMAP_LOOKUP[change["new"]])
                self.substitute_layer(self.current_basemap, new_basemap)
                self.current_basemap = new_basemap

        # def on_toggle_change(change):
        #     """Toggle visibility of dropdown"""
        #     if change["new"]:
        #         dropdown.layout.visibility = 'visible'  # Show dropdown
        #     else:
        #         dropdown.layout.visibility = 'hidden'   # Hide dropdown

        # Set up event handlers
        dropdown.observe(on_dropdown_change, names="value")

        # Assemble and add control
        widget_box = HBox([toggle, dropdown])

        is_open = [True]

        def on_toggle_change(change):
            """Toggle visibility of dropdown"""
            if is_open[0] and change["new"]:
                widget_box.children = (toggle, dropdown)
            else:
                widget_box.children = (toggle,)

        toggle.observe(on_toggle_change, names="value")

        control = WidgetControl(widget=widget_box, position="topright")
        self.add_control(control)

    def add_search_marker(self, **kwargs):
        """Adds a search marker to the map."""

        from ipyleaflet import Map, SearchControl, Marker, AwesomeIcon

        search_marker = Marker(
            icon=AwesomeIcon(name="search", marker_color="blue", icon_color="white")
        )

        search_control = SearchControl(
            position="topleft",
            url="https://nominatim.openstreetmap.org/search?format=json&q={s}",
            zoom=10,
            marker=search_marker,
        )

        self.add(search_control)



import requests
import ipywidgets as widgets
import geopandas as gpd
import pandas as pd
from shapely.geometry import Point
import shapely
from pyproj import Transformer
import os
import numpy as np

# ee.Authenticate()
# ee.Initialize(project="ee-forestplotvariables")


class buffer:
    def __init__(self):
        """
        Initializes the buffer_coordinates class and sets up the GUI components.
        Part 1 of buffered coordinates approach which creates circles around points.
        (Part 2 is in buffer_method.py)
        """
        # File Upload
        self.file_upload = widgets.FileUpload(
            accept=".csv, .txt",  # Accepted file extension e.g. '.txt', '.pdf', 'image/*', 'image/*,.pdf'
            multiple=False,  # True to accept multiple files upload else False
        )

        # Dropdown
        url = "https://raw.githubusercontent.com/opengeos/geospatial-data-catalogs/master/gee_catalog.json"
        data = self.fetch_geojson(url)
        data_dict = {item["title"]: item["id"] for item in data if "title" in item}
        self.dropdown = widgets.Dropdown(
            options=data_dict,  # keys shown, values returned
            description="OPTIONAL:",
            disabled=False,
        )
        self.dropdown.observe(self.on_dropdown_change, names="value")

        self.buffer_radius = widgets.FloatText(
            value=100,
            description="Buffer radius (ft):",
            disabled=False,
            display="flex",
            flex_flow="column",
            align_items="stretch",
            style={"description_width": "initial"},
        )

        self.sample_button = widgets.IntText(
            value=10,
            description="Sampling number (int):",
            disabled=False,
            display="flex",
            flex_flow="column",
            align_items="stretch",
            style={"description_width": "initial"},
        )

        self.run_button = widgets.Button(
            description="Run Query",
            disabled=False,
            button_style="",  # 'success', 'info', 'warning', 'danger' or ''
            tooltip="Click me",
            icon="rotate right",  # (FontAwesome names without the `fa-` prefix)
        )

        self.output = widgets.Output()

        self.run_button.on_click(self.on_button_clicked)

        self.hbox = widgets.HBox(
            [
                self.file_upload,
                self.buffer_radius,
                self.sample_button,
                self.run_button,
            ]
        )

        self.vbox = widgets.VBox([self.hbox, self.output])

    def on_button_clicked(self, b):
        """
        Callback function to handle button click events."""
        with self.output:
            self.output.clear_output()
            print(
                f"GeoJSON file will be saved to Downloads folder under this name:{self.buffer_radius.value}ft.csv"
            )

            import io

            if self.file_upload.value:
                # For the first file (if multiple=False)
                file_info = self.file_upload.value[0]
                content_bytes = file_info["content"].tobytes()  # file content as bytes
                points = pd.read_csv(io.BytesIO(content_bytes))
                lat_cols = ["lat", "latitude", "y", "LAT", "Latitude", "Lat", "Y"]
                lon_cols = [
                    "lon",
                    "long",
                    "longitude",
                    "x",
                    "LON",
                    "Longitude",
                    "Long",
                    "X",
                ]
                id_cols = ["id", "ID", "plot_ID", "plot_id", "plotID", "plotId"]

                def find_column(possible_names, columns):
                    for name in possible_names:
                        if name in columns:
                            return name
                    # fallback: check case-insensitive match
                    lower_columns = {c.lower(): c for c in columns}
                    for name in possible_names:
                        if name.lower() in lower_columns:
                            return lower_columns[name.lower()]
                    raise ValueError(f"No matching column found for {possible_names}")

                lat_col = find_column(lat_cols, points.columns)
                lon_col = find_column(lon_cols, points.columns)
                id_col = find_column(id_cols, points.columns)
                points = points.rename(
                    columns={lat_col: "LAT", lon_col: "LON", id_col: "plot_ID"}
                )
            else:
                print("Please upload a CSV file.")

            radius_ft = self.buffer_radius.value
            no_samp = self.sample_button.value

            out_dir = os.path.join(os.path.expanduser("~"), "Downloads")
            output_file = f"{radius_ft}ft.csv"
            out_path = os.path.join(out_dir, output_file)

            self.obfuscate_points(
                data=points,
                radius_feet=radius_ft,
                no_samp=no_samp,
                plot_id_col="plot_ID",
                output_file=out_path,
            )
            print(f"Buffered coordinates saved to {out_path}")

    def on_dropdown_change(self, change):
        """
        Callback function to handle dropdown value changes.
        """
        if change["new"]:
            with self.output:
                self.output.clear_output()
                catalog = "https://raw.githubusercontent.com/opengeos/geospatial-data-catalogs/master/gee_catalog.json"
                data = self.fetch_geojson(catalog)
                data_dict = {item["id"]: item["url"] for item in data if "id" in item}
                change_value = str(change["new"])
                url = data_dict.get(change_value)
                print(f"Selected dataset: {change['new']}")
                print(f"URL: {url}")

    def create_obfuscated_points(self, point, radius_feet, no_samp, crs="EPSG:4326"):
        """
        Create a circle polygon (as a shapely geometry) with the given radius in feet,
        where the provided point is randomly located inside the circle (not at the center).
        """
        # Convert radius from feet to meters
        radius_m = radius_feet * 0.3048

        # Project to local UTM for accurate distance calculations
        utm_crs = f"EPSG:326{int((point.x + 180) // 6) + 1}"
        transformer_to_utm = Transformer.from_crs(crs, utm_crs, always_xy=True)
        transformer_to_latlon = Transformer.from_crs(utm_crs, crs, always_xy=True)
        x, y = transformer_to_utm.transform(point.x, point.y)

        # Randomize the point's location within the circle
        points = []
        for no in range(no_samp):
            # Randomize the point's location within the circle
            angle = np.random.uniform(0, 2 * np.pi)
            distance = np.random.uniform(0, radius_m)
            # Calculate center of the circle so that the point is inside the circle but not at the center
            center_x = x - distance * np.cos(angle)
            center_y = y - distance * np.sin(angle)
            center = Point(center_x, center_y)
            center_latlon = shapely.ops.transform(
                lambda x, y: transformer_to_latlon.transform(x, y), center
            )
            points.append(center_latlon)
        return points

    def obfuscate_points(self, data, radius_feet, no_samp, plot_id_col, output_file):
        """
        Obfuscate points within a radius and save as csv.

        Args:
            data (str, pd.DataFrame, gpd.GeoDataFrame): Input data (GeoJSON, DataFrame, or GeoDataFrame).
            radius_feet (float): Radius of the circle in feet.
            plot_id_col (str): Column name for plot IDs.
            output_file (str): Path to save the output GeoJSON file.
        """
        if isinstance(data, str):
            coordinates = pd.read_csv(data)
            gdf = gpd.GeoDataFrame(
                coordinates,
                geometry=gpd.points_from_xy(coordinates.LON, coordinates.LAT),
                crs="EPSG:4326",  # Directly set CRS during creation
            )
        elif isinstance(data, pd.DataFrame):
            coordinates = data
            gdf = gpd.GeoDataFrame(
                coordinates,
                geometry=gpd.points_from_xy(coordinates.LON, coordinates.LAT),
                crs="EPSG:4326",  # Directly set CRS during creation
            )
        else:
            gdf = data.to_crs(epsg=4326)  # Ensure WGS84

        centers = []

        for idx, row in gdf.iterrows():
            point = row["geometry"]
            center = self.create_obfuscated_points(
                point, radius_feet, no_samp, crs=gdf.crs
            )
            for pt in center:
                centers.append({"plot_ID": row[plot_id_col], "lat": pt.y, "lon": pt.x})
            # Create new GeoDataFrame
        df = pd.DataFrame(centers)
        # point_df = pd.DataFrame(gdf_circles)
        point_csv = df.to_csv()
        with open(output_file, "w", newline="") as f:
            f.write(point_csv)
        print(f"CSV saved to {output_file}")
        return point_csv

    def fetch_geojson(self, url):
        """
        Fetch GeoJSON data from a given URL.

        Args:
            url (str): URL to fetch the GeoJSON data from.
        """
        try:
            response = requests.get(url)
            response.raise_for_status()  # Raises an exception for HTTP errors
            geojson_data = response.json()  # Parse the JSON response
            return geojson_data
        except requests.exceptions.HTTPError as http_err:
            print(f"HTTP error occurred: {http_err}")
        except requests.exceptions.ConnectionError as conn_err:
            print(f"Error connecting to the server: {conn_err}")
        except Exception as err:
            print(f"An error occurred: {err}")
        return None



import requests
import ipywidgets as widgets
import geopandas as gpd
import pandas as pd
import geemap as gm
import ee
import os

# ee.Authenticate()
# ee.Initialize(project="ee-forestplotvariables")


class point_extraction:
    def __init__(self):
        # File Upload
        self.file_upload = widgets.FileUpload(
            accept=".csv, .txt",  # Accepted file extension e.g. '.txt', '.pdf', 'image/*', 'image/*,.pdf'
            multiple=False,  # True to accept multiple files upload else False
        )
        # Dropdown
        url = "https://raw.githubusercontent.com/opengeos/geospatial-data-catalogs/master/gee_catalog.json"
        data = point_extraction.fetch_geojson(url)
        data_dict = {item["title"]: item["id"] for item in data if "title" in item}
        self.dropdown = widgets.Dropdown(
            options=data_dict,  # keys shown, values returned
            description="Dataset:",
            disabled=False,
        )
        self.start_date = widgets.DatePicker(description="Start Date", disabled=False)
        self.end_date = widgets.DatePicker(description="End Date", disabled=False)
        self.run_button = widgets.Button(
            description="Run Query",
            disabled=False,
            button_style="",  # 'success', 'info', 'warning', 'danger' or ''
            tooltip="Click me",
            icon="rotate right",  # (FontAwesome names without the `fa-` prefix)
        )
        self.output = widgets.Output()
        self.run_button.on_click(self.on_button_clicked)
        self.dropdown.observe(self.on_dropdown_change, names="value")
        self.hbox = widgets.HBox(
            [
                self.file_upload,
                self.dropdown,
                self.start_date,
                self.end_date,
                self.run_button,
            ]
        )
        self.vbox = widgets.VBox([self.hbox, self.output])

    def on_dropdown_change(self, change):
        if change["new"]:
            with self.output:
                self.output.clear_output()
                catalog = "https://raw.githubusercontent.com/opengeos/geospatial-data-catalogs/master/gee_catalog.json"
                data = point_extraction.fetch_geojson(catalog)
                data_dict = {item["id"]: item["url"] for item in data if "id" in item}
                change_value = str(change["new"])
                url = data_dict.get(change_value)
                print(f"Selected dataset: {change['new']}")
                print(f"URL: {url}")

    def on_button_clicked(self, b):
        with self.output:
            self.output.clear_output()
            print(
                f"You entered: {self.dropdown.value}. CSV file will be saved to Downloads folder under this name."
            )
            import io

            if self.file_upload.value:
                # For the first file (if multiple=False)
                file_info = self.file_upload.value[0]
                content_bytes = file_info["content"].tobytes()  # file content as bytes
                points = pd.read_csv(io.BytesIO(content_bytes))
                lat_cols = ["lat", "latitude", "y", "LAT", "Latitude", "Lat", "Y"]
                lon_cols = [
                    "lon",
                    "long",
                    "longitude",
                    "x",
                    "LON",
                    "Longitude",
                    "Long",
                    "X",
                ]
                id_cols = ["id", "ID", "plot_ID", "plot_id", "plotID", "plotId"]

                def find_column(possible_names, columns):
                    for name in possible_names:
                        if name in columns:
                            return name
                    # fallback: check case-insensitive match
                    lower_columns = {c.lower(): c for c in columns}
                    for name in possible_names:
                        if name.lower() in lower_columns:
                            return lower_columns[name.lower()]
                    raise ValueError(f"No matching column found for {possible_names}")

                lat_col = find_column(lat_cols, points.columns)
                lon_col = find_column(lon_cols, points.columns)
                id_col = find_column(id_cols, points.columns)
                points = points.rename(
                    columns={lat_col: "LAT", lon_col: "LON", id_col: "plot_ID"}
                )
            else:
                print("Please upload a CSV file.")
            geedata = self.dropdown.value
            start_date = self.start_date.value
            end_date = self.end_date.value
            self.get_coordinate_data(
                data=points, geedata=geedata, start_date=start_date, end_date=end_date
            )

    def get_coordinate_data(self, data, geedata, start_date, end_date, **kwargs):
        """
        Pull data from provided coordinates from GEE.

        Args:
            data (str): The data to get the coordinate data from.

        Returns:
            data (str): CSV file contained GEE data.
        """

        # Load data with safety checks
        if isinstance(data, str):
            coordinates = pd.read_csv(data)
            gdf = gpd.GeoDataFrame(
                coordinates,
                geometry=gpd.points_from_xy(coordinates.LON, coordinates.LAT),
                crs="EPSG:4326",  # Directly set CRS during creation
            )
        elif isinstance(data, pd.DataFrame):
            coordinates = data
            gdf = gpd.GeoDataFrame(
                coordinates,
                geometry=gpd.points_from_xy(coordinates.LON, coordinates.LAT),
                crs="EPSG:4326",  # Directly set CRS during creation
            )
        else:
            gdf = data.to_crs(epsg=4326)  # Ensure WGS84
        geojson = gdf.__geo_interface__
        fc = gm.geojson_to_ee(geojson)
        # Load the GEE dataset as an image
        geeimage = point_extraction.load_gee_as_image(geedata, start_date, end_date)
        name = f"{geedata}"
        file_name = name.replace("/", "_")
        out_dir = os.path.join(os.path.expanduser("~"), "Downloads")
        output_file = f"{file_name}.csv"
        out_path = os.path.join(out_dir, output_file)
        # Retrieve data from the image using sampleRegions
        sampled_data = gm.extract_values_to_points(fc, geeimage, out_path)
        return sampled_data

    def fetch_geojson(url):
        try:
            response = requests.get(url)
            response.raise_for_status()  # Raises an exception for HTTP errors
            geojson_data = response.json()  # Parse the JSON response
            return geojson_data
        except requests.exceptions.HTTPError as http_err:
            print(f"HTTP error occurred: {http_err}")
        except requests.exceptions.ConnectionError as conn_err:
            print(f"Error connecting to the server: {conn_err}")
        except Exception as err:
            print(f"An error occurred: {err}")
        return None

    def create_dropdown():
        """
        Creates an ipywidgets dropdown menu from a GeoJSON catalog.

        Args:
            url (str, optional): URL to the GeoJSON catalog. Defaults to the Opengeos catalog.

        Returns:
            ipywidgets.Dropdown: A dropdown widget with the names from the catalog.
        """

        url = "https://raw.githubusercontent.com/opengeos/geospatial-data-catalogs/master/gee_catalog.json"
        data = point_extraction.fetch_geojson(url)
        data_dict = {item["title"]: item["id"] for item in data if "title" in item}
        dropdown = widgets.Dropdown(
            options=data_dict,  # keys shown, values returned
            description="Dataset:",
            disabled=False,
        )
        return dropdown

    def add_date_picker():
        date_picker = widgets.DatePicker(description="Pick a Date", disabled=False)
        return date_picker

    def load_gee_as_image(dataset_id, start_date, end_date, **kwargs):
        """
        Loads any GEE dataset (Image, ImageCollection, FeatureCollection) as an ee.Image.
        Optionally filters by start and end date if applicable.

        Parameters:
            dataset_id (str): The Earth Engine dataset ID.
            start_date (str): Optional start date in 'YYYY-MM-DD' format.
            end_date (str): Optional end date in 'YYYY-MM-DD' format.

        Returns:
            ee.Image: The resulting image.
        """
        url = "https://raw.githubusercontent.com/opengeos/geospatial-data-catalogs/master/gee_catalog.json"
        response = requests.get(url)
        response.raise_for_status()  # Raises an exception for HTTP errors
        geojson_data = response.json()
        data_type = [item["type"] for item in geojson_data if item["id"] == dataset_id]
        data_str = " ".join(data_type)
        start_date = str(start_date)
        end_date = str(end_date)
        # Try loading as Image
        if data_str == "image":
            img = ee.Image(dataset_id)
            # If .getInfo() doesn't throw, it's an Image
            img.getInfo()
            return img
        elif data_str == "image_collection":
            col = ee.ImageCollection(dataset_id)
            # If date filters are provided, apply them
            if start_date is None and end_date is None:
                col = col.filterDate(start_date, end_date)
            else:
                pass
            # Reduce to a single image (e.g., median composite)
            img = col.median()
            return img
        # Try loading as FeatureCollection (convert to raster)
        else:
            # fc_temp = ee.FeatureCollection(dataset_id)
            # if start_date is None and end_date is None:
            #         fc_temp = fc_temp.filterDate(start_date, end_date)
            # # Convert to raster: burn a value of 1 into a new image
            # img = fc_temp.reduceToImage(properties=[], reducer=ee.Reducer.median())
            # img.getInfo()
            # return img
            # or print(f"Dataset must be either an Image or Image Collection")
            raise ValueError("Dataset ID is not a valid Image or ImageCollection.")
